# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from ......schema_classes import ChartKeyClass
from ......schema_classes import CorpGroupKeyClass
from ......schema_classes import CorpUserKeyClass
from ......schema_classes import DashboardKeyClass
from ......schema_classes import DataFlowKeyClass
from ......schema_classes import DataHubPolicyKeyClass
from ......schema_classes import DataJobKeyClass
from ......schema_classes import DataPlatformKeyClass
from ......schema_classes import DataProcessKeyClass
from ......schema_classes import DatasetKeyClass
from ......schema_classes import GlossaryNodeKeyClass
from ......schema_classes import GlossaryTermKeyClass
from ......schema_classes import MLFeatureKeyClass
from ......schema_classes import MLFeatureTableKeyClass
from ......schema_classes import MLModelDeploymentKeyClass
from ......schema_classes import MLModelGroupKeyClass
from ......schema_classes import MLModelKeyClass
from ......schema_classes import MLPrimaryKeyKeyClass
from ......schema_classes import TagKeyClass


ChartKey = ChartKeyClass
CorpGroupKey = CorpGroupKeyClass
CorpUserKey = CorpUserKeyClass
DashboardKey = DashboardKeyClass
DataFlowKey = DataFlowKeyClass
DataHubPolicyKey = DataHubPolicyKeyClass
DataJobKey = DataJobKeyClass
DataPlatformKey = DataPlatformKeyClass
DataProcessKey = DataProcessKeyClass
DatasetKey = DatasetKeyClass
GlossaryNodeKey = GlossaryNodeKeyClass
GlossaryTermKey = GlossaryTermKeyClass
MLFeatureKey = MLFeatureKeyClass
MLFeatureTableKey = MLFeatureTableKeyClass
MLModelDeploymentKey = MLModelDeploymentKeyClass
MLModelGroupKey = MLModelGroupKeyClass
MLModelKey = MLModelKeyClass
MLPrimaryKeyKey = MLPrimaryKeyKeyClass
TagKey = TagKeyClass
# fmt: on
